// Env.cs
using System;
using System.Collections;


namespace TtLisp {
  public class Env {
    // -- instance -------------------------------------------------------
    private Stack         symbolStack;
    private FunctionTable functionTable;

    public Env() {
      this.symbolStack = new Stack();
      this.symbolStack.Push( new SymbolTable() );

      this.SymbolTable.ConstIntern( Nil.NIL, Nil.NIL );
      this.SymbolTable.ConstIntern( True.T, True.T );

      this.functionTable = new FunctionTable();

      this.DefineFunction( "car",           new SystemFunctions.FunctionCar() );
      this.DefineFunction( "cdr",           new SystemFunctions.FunctionCdr() );
      this.DefineFunction( "quote",         new SystemFunctions.FunctionQuote() );
      this.DefineFunction( "cons",          new SystemFunctions.FunctionCons() );
      this.DefineFunction( "atom",          new SystemFunctions.FunctionAtom() );
      this.DefineFunction( "eq",            new SystemFunctions.FunctionEq() );
      this.DefineFunction( "progn",         new SystemFunctions.FunctionProgn() );
      this.DefineFunction( "if",            new SystemFunctions.FunctionIf() );
      this.DefineFunction( "cond",          new SystemFunctions.FunctionCond() );
      this.DefineFunction( "lambda",        new SystemFunctions.FunctionLambda() );
      this.DefineFunction( "macro",         new SystemFunctions.FunctionMacro() );
      this.DefineFunction( "bind-function", new SystemFunctions.FunctionBindFunction() );
      this.DefineFunction( "function",      new SystemFunctions.FunctionFunction() );
      this.DefineFunction( "apply",         new SystemFunctions.FunctionApply() );
      this.DefineFunction( "funcall",       new SystemFunctions.FunctionFuncall() );
      this.DefineFunction( "set",           new SystemFunctions.FunctionSet() );
      this.DefineFunction( "setq",          new SystemFunctions.FunctionSetQ() );
      this.DefineFunction( "eval",          new SystemFunctions.FunctionEval() );
      this.DefineFunction( "list",          new SystemFunctions.FunctionList() );
      this.DefineFunction( "macroexpand",   new SystemFunctions.FunctionMacroExpand() );
      this.DefineFunction( "let*",           new SystemFunctions.FunctionLetAste() );

      this.DefineFunction( "defun",
                           new Macro( List.MakeAsSymbol( "_name",
                                                         "_lambda-list",
                                                         "_body" ),
                                      List.Make( new Symbol( "list" ),
                                                 List.MakeAsSymbol( "quote", "bind-function" ),
                                                 List.Make( new Symbol("list" ),
                                                            List.MakeAsSymbol( "quote", "quote" ),
                                                            new Symbol( "_name" ) ),
                                                 List.Make( new Symbol( "list" ),
                                                            List.MakeAsSymbol( "quote", "lambda" ),
                                                            new Symbol( "_lambda-list" ),
                                                            new Symbol( "_body" ) ) ) ) );

      this.DefineFunction( "defmacro",
                           new Macro( List.MakeAsSymbol( "_name",
                                                         "_lambda-list",
                                                         "_body" ),
                                      List.Make( new Symbol( "list" ),
                                                 List.MakeAsSymbol( "quote", "bind-function" ),
                                                 List.Make( new Symbol("list" ),
                                                            List.MakeAsSymbol( "quote", "quote" ),
                                                            new Symbol( "_name" ) ),
                                                 List.Make( new Symbol( "list" ),
                                                            List.MakeAsSymbol( "quote", "macro" ),
                                                            new Symbol( "_lambda-list" ),
                                                            new Symbol( "_body" ) ) ) ) );

    }

    private void DefineFunction( string name, Function func ) {
      this.FunctionTable.Define( new Symbol( name ), func );
    }

    public SymbolicExpression Eval( SymbolicExpression s ) {
      return s.ToEvaluated( this );
    }

    private Stack SymbolStack { get { return this.symbolStack; } }
    public SymbolTable SymbolTable {
      get { return (SymbolTable)this.SymbolStack.Peek(); }
    }
    public void PushStack() {
      this.SymbolStack.Push( new SymbolTable( this.SymbolTable ) );
    }
    public void PopStack() {
      this.SymbolStack.Pop();
    }

    public FunctionTable FunctionTable {
      get { return this.functionTable; }
    }

  }
}
