// List.cs
using System;
using System.Text;


namespace TtLisp {
  public class List : SymbolicExpression {
    // -- static ---------------------------------------------------------
    public static List Make( params SymbolicExpression[] args ) {
      List list = Nil.NIL;
      for ( int i = args.Length - 1; i >= 0; --i ) {
        list = new List( args[i], list );
      }
      return list;
    }

    public static List MakeAsSymbol( params string[] args ) {
      List list = Nil.NIL;
      for ( int i = args.Length - 1; i >= 0; --i ) {
        list = new List( new Symbol( args[i] ), list );
      }
      return list;
    }

    // -- instance -------------------------------------------------------
    private readonly SymbolicExpression car;
    private readonly SymbolicExpression cdr;

    public List( SymbolicExpression car ) : this( car, Nil.NIL ) {}
    public List( SymbolicExpression car, SymbolicExpression cdr ) {
      this.car = car;
      this.cdr = cdr;
    }

    public SymbolicExpression Car {
      get { return this.car; }
    }

    public SymbolicExpression Cdr {
      get { return this.cdr; }
    }

    public SymbolicExpression this[int index] {
      get {
        List cur = this;
        for ( int i = 0; i < index; ++i ) {
          cur = (List)cur.Cdr;
        }
        return cur.Car;
      }
    }

    public int Size {
      get {
        if ( this is Nil ) { return 0; }
        if ( !(this.Cdr is List) ) { return 1; }
        return ((List)this.Cdr).Size + 1;
      }
    }

    public override SymbolicExpression ToEvaluated( Env env ) {
      Function func;
      if ( this.Car is ISymbol ) {
        ISymbol _tmp = (ISymbol)this.Car;
        if ( env.FunctionTable.ContainsKey( _tmp ) ) {
          func = env.FunctionTable[_tmp];
        }
        else {
          throw new FunctionIsNotDefinedException();
        }
      }
      else {
        SymbolicExpression _tmp = env.Eval( this.Car );
        Console.WriteLine( this );
        if ( _tmp is Function ) {
          func = (Function)_tmp;
        }
        else {
          throw new InvalidFunctionException();
        }
      }
      return func.Call( env, this.Cdr is List ? (List)this.Cdr : Nil.NIL );
    }

    public override string ToString() {
      if ( this.Size == 2 && this.Car is ISymbol && ((ISymbol)this.Car).Name == "quote" ) {
        return String.Concat( "'", this[1].ToString() );
      }

      StringBuilder sb = new StringBuilder();
      sb.Append( "(" );
      for ( List cur = this; ; cur = (List)cur.Cdr ) {
        sb.Append( cur.Car.ToString() );
        if ( cur.Cdr is Nil ) { break; }
        if ( cur.Cdr is Atom ) {
          sb.Append( " . " );
          sb.Append( cur.Cdr.ToString() );
          break;
        }
        sb.Append( " " );
      }
      sb.Append( ")" );
      return sb.ToString();
    }

  }
}
