// Numeric.cs
using System;

namespace TtLisp {
  public abstract class Numeric : Atom {}

  public class Integer : Numeric {
    // -- instance -------------------------------------------------------
    private int val;

    public Integer() : this( 0 ) {}
    public Integer( int i )  { this.val = i; }

    public int Value { get { return this.val; } }

    public Integer Add( Integer other ) {
      return new Integer( this.val + other.val );
    }
    public Integer Sub( Integer other ) {
      return new Integer( this.val - other.val );
    }
    public Integer Mul( Integer other ) {
      return new Integer( this.val * other.val );
    }
    public Integer Div( Integer other ) {
      return new Integer( this.val / other.val );
    }
    public Integer Mod( Integer other ) {
      return new Integer( this.val % other.val );
    }

    public SymbolicExpression Eq( Integer other ) {
      return this.val == other.val ? True.T : SymbolicExpression.NIL;
    }
    public SymbolicExpression Ne( Integer other ) {
      return this.val != other.val ? True.T : SymbolicExpression.NIL;
    }
    public SymbolicExpression Gt( Integer other ) {
      return this.val > other.val ? True.T : SymbolicExpression.NIL;
    }
    public SymbolicExpression Lt( Integer other ) {
      return this.val < other.val ? True.T : SymbolicExpression.NIL;
    }
    public SymbolicExpression Ge( Integer other ) {
      return this.val >= other.val ? True.T : SymbolicExpression.NIL;
    }
    public SymbolicExpression Le( Integer other ) {
      return this.val <= other.val ? True.T : SymbolicExpression.NIL;
    }

    protected override bool LispEqThisType( SymbolicExpression other ) {
      return other is Integer && this.val.Equals( ((Integer)other).val );
    }

    public override string ToString() {
      return this.val.ToString();
    }

  }
}
