// SymbolTable.cs
using System;
using System.Collections;

namespace TtLisp {
  public class SymbolTable : IEnumerable {
    private Hashtable self;
    private Hashtable consts;
    private SymbolTable top;

    public SymbolTable() {
      this.self = new Hashtable();
      this.consts = new Hashtable();
      this.top = this;
    }
    public SymbolTable( SymbolTable other ) : this() {
      this.self = new Hashtable();
      this.consts = other.top.consts;
      this.top = other.top;
    }

    public virtual int Count { get { return self.Count; } }
    public virtual bool IsSynchronized { get { return self.IsSynchronized; } }

    public virtual SymbolicExpression this[ISymbol key] {
      get {
        if ( this.ContainsKey( key ) ) {
          return (SymbolicExpression)self[key.Name];
        }
        else if ( this.top.ContainsKey( key ) ) {
          return (SymbolicExpression)this.top.self[key.Name];
        }
        throw new SymbolUnboundException();
      }
    }

    public virtual ICollection Keys { get { return self.Keys; } }
    public virtual object SyncRoot { get { return self.SyncRoot; } }
    public virtual ICollection Values { get { return self.Values; } }

    public virtual void Intern( ISymbol key, SymbolicExpression value ) {
      if ( this.consts.ContainsKey( key ) ) {
        throw new ConstYetDefinedException();
      }
      ( this.ContainsKey( key ) ? this : this.top ).LocalIntern( key, value );
    }

    public virtual void LocalIntern( ISymbol key, SymbolicExpression value ) {
      if ( this.consts.ContainsKey( key ) ) {
        throw new ConstYetDefinedException();
      }
      self[key.Name] = value;
    }

    public virtual void ConstIntern( ISymbol key, SymbolicExpression value ) {
      if ( this.consts.ContainsKey( key ) ) {
        throw new ConstYetDefinedException();
      }
      this.top.LocalIntern( key, value );
      this.consts.Add( key.Name, true );
    }

    public virtual bool ContainsKey( ISymbol key ) {
      return self.ContainsKey( key.Name );
    }
    public virtual bool ContainsValue( SymbolicExpression value ) {
      return self.ContainsValue( value );
    }
    public virtual IEnumerator GetEnumerator() {
      return self.GetEnumerator();
    }

    public virtual void Unintern( ISymbol key ) {
      if ( this.consts.ContainsKey( key ) ) {
        throw new ConstCanNotUndefineException();
      }
      ( this.ContainsKey( key ) ? self : this.top.self ).Remove( key.Name );
    }
  }
}
