using System;
using TtLisp;

class Test {
  static public void p( object o ) { Console.WriteLine( o ); }
  static public void p( string fmt, params object[] o ) { Console.WriteLine( fmt, o ); }
  static public int Main( string[] args ) {
    env = new Env();

    List list;
    list = MakeList( new Symbol( "quote" ), MakeList( new Integer( 100 ), new LString( "amano" ) ) );
    PrintAndEval( list );
    Funcall( "car", list );
    Funcall( "cdr", list );

    Funcall( "cons", list, MakeList( new Symbol( "quote" ), new Symbol( "amano" ) ) );

    Funcall( "atom", Nil.NIL );
    Funcall( "atom", list );
    Funcall( "atom", MakeList( new Symbol( "car" ), list ) );

    Funcall( "setq", new Symbol( "x" ), list );
    Funcall( "setq", new Symbol( "_tmp" ), list );
    Funcall( "set", MakeList( new Symbol( "quote" ), new Symbol( "x" ) ), list );

    PrintAndEval( new Symbol( "x" ) );
    Funcall( "eq", new Symbol( "x" ), new Symbol( "x" ) );
    Funcall( "eq", new Symbol( "x" ), new Symbol( "_tmp" ) );
    Funcall( "eq", new Integer( 100 ), new Symbol( "x" ) );
    Funcall( "eq", new Integer( 500 ), new Integer( 500 ) );

    Funcall( "progn", new Integer( 10 ), new Integer( 20 ), new Integer( 30 ) );

    List cond;
    cond = MakeList( new Symbol( "atom" ), Nil.NIL );
    Funcall( "if", cond, new Symbol( "x" ) );
    cond = MakeList( new Symbol( "atom" ), list );
    Funcall( "if", cond, new Symbol( "x" ) );
    Funcall( "if", cond, new Symbol( "x" ), True.T );

    Funcall( "cond",
             MakeList( MakeList( new Symbol( "atom" ), Nil.NIL ), new Integer( 100 ) ),
             MakeList( MakeList( new Symbol( "atom" ), True.T ), new Integer( 200 ) ) );

    {
      List llist, body;
      llist = MakeList( new Symbol( "x" ), new Symbol( "y" ), new Symbol( "z" ) );
      body  = MakeList( new Symbol( "cons" ), new Symbol( "x" ), MakeList( new Symbol( "cons" ), new Symbol( "y" ), new Symbol( "z" ) ) );
      Funcall( "lambda", llist, body );
      Funcall( "bind-function", MakeList( new Symbol( "quote" ), new Symbol( "foo" ) ), MakeList( new Symbol( "lambda" ), llist, body ) );
      Funcall( "foo", new Integer( 10 ), new Integer( 20 ), new Integer( 30 ) );
      Funcall( "foo", new Symbol( "x" ), new Symbol( "x" ), new Symbol( "x" ) );

      llist = MakeList( new Symbol( "n" ) );
      body = MakeList( new Symbol( "list" ) , MakeList( new Symbol( "quote" ), new Symbol( "cons" ) ), new Symbol( "n" ), new Integer( 1024 ) );
      Funcall( "macro", llist, body );
      Funcall( "bind-function", MakeList( new Symbol( "quote" ), new Symbol( "bar" ) ), MakeList( new Symbol( "macro" ), llist, body ) );
      Funcall( "bar", new Integer( 10 ) );
      Funcall( "bar", new Symbol( "x" ) );
    }
    Funcall( "function", new Symbol( "foo" ) );
    Funcall( "function", new Symbol( "bar" ) );
    Funcall( "function", new Symbol( "car" ) );

    Funcall( "set", MakeList( new Symbol( "quote" ), new Symbol( "y" ) ), MakeList( new Symbol( "quote" ), MakeList( new Symbol( "cons" ), new Integer( 888 ), new Integer( 777 ) ) ) );
    Funcall( "eval", new Symbol( "y" ) );

    Funcall( "list", new Integer( 256 ), new Symbol( "x" ), list );

    Funcall( "apply",
             MakeList( new Symbol( "quote" ), new Symbol( "car" ) ),
             MakeList( new Symbol( "quote" ), MakeList( MakeList( new Symbol( "a" ),
                                                                  new Symbol( "b" ),
                                                                  new Symbol( "c" ) ) ) ) );
    Funcall( "funcall",
             MakeList( new Symbol( "quote" ), new Symbol( "car" ) ),
             MakeList( new Symbol( "quote" ), MakeList( new Symbol( "a" ),
                                                        new Symbol( "b" ),
                                                        new Symbol( "c" ) ) ) );


    Funcall( "macroexpand", MakeList( new Symbol( "quote" ),
                                      List.MakeAsSymbol( "bar", "x" ) ) );

    Funcall( "function", new Symbol( "defun" ) );
    Funcall( "function", new Symbol( "defmacro" ) );


    Funcall( "macroexpand", MakeList( new Symbol( "quote" ),
                                      MakeList( new Symbol( "defun" ),
                                                new Symbol( "baz" ),
                                                List.MakeAsSymbol( "x", "y" ),
                                                List.MakeAsSymbol( "cons", "x", "y" ) ) ) );
    
    Funcall( "defun",
             new Symbol( "baz" ),
             List.MakeAsSymbol( "x", "y" ),
             List.MakeAsSymbol( "cons", "x", "y" ) );
    
    Funcall( "baz", new Integer( 1234 ), new LString( "amano" ));

    Funcall( "let*",
             List.Make( new Symbol( "y" ), List.Make( new Symbol( "x" ), new Integer( 789 ) ), new Symbol( "z" ) ),
             List.MakeAsSymbol( "setq", "z", "x" ),
             List.MakeAsSymbol( "list", "z", "y", "x" ) );

    return 0;
  }

  static Env env;

  static SymbolicExpression PrintAndEval( SymbolicExpression s ) {
    p( s );
    SymbolicExpression e_s = env.Eval( s );
    p( "-> {0}", e_s );
    return e_s;
  }

  static List MakeList( params SymbolicExpression[] args ) {
    List list = Nil.NIL;
    for ( int i = args.Length - 1; i >= 0; --i ) {
      list = new List( args[i], list );
    }
    return list;
  }

  static SymbolicExpression Funcall( string name, params SymbolicExpression[] args ) {
    SymbolicExpression[] tmp = new SymbolicExpression[args.Length + 1];
    tmp[0] = new Symbol( name );
    args.CopyTo( tmp, 1 );
    return PrintAndEval( MakeList( tmp ) );
  }
}
